from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional

from fastapi import APIRouter, Depends, Request
from fastapi.responses import FileResponse
from pydantic import BaseModel
from sqlmodel import select

import murfey.server.api.websocket as ws
from murfey.server import _transport_object
from murfey.server.api import templates
from murfey.server.api.auth import (
    MurfeyInstrumentNameFrontend as MurfeyInstrumentName,
    MurfeySessionIDFrontend as MurfeySessionID,
    validate_token,
)
from murfey.server.api.session_shared import (
    find_upstream_visits as _find_upstream_visits,
    gather_upstream_files as _gather_upstream_files,
    gather_upstream_tiffs as _gather_upstream_tiffs,
    get_foil_hole as _get_foil_hole,
    get_foil_holes_from_grid_square as _get_foil_holes_from_grid_square,
    get_grid_squares as _get_grid_squares,
    get_grid_squares_from_dcg as _get_grid_squares_from_dcg,
    get_tiff_file as _get_tiff_file,
    get_upstream_file as _get_upstream_file,
    remove_session_by_id,
)
from murfey.server.ispyb import DB as ispyb_db, get_all_ongoing_visits
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise
from murfey.util.config import get_machine_config
from murfey.util.db import (
    ClassificationFeedbackParameters,
    ClientEnvironment,
    DataCollection,
    DataCollectionGroup,
    FoilHole,
    GridSquare,
    Movie,
    ProcessingJob,
    RsyncInstance,
    Session,
    SessionProcessingParameters,
    SPARelionParameters,
    Tilt,
    TiltSeries,
)
from murfey.util.models import UpstreamFileRequestInfo, Visit

logger = getLogger("murfey.server.api.session_info")

router = APIRouter(
    prefix="/session_info",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: General"],
)


@router.get("/health/")
def health_check(db=ispyb_db):
    conn = db.connection()
    conn.close()
    return {
        "ispyb_connection": True,
        "rabbitmq_connection": _transport_object.transport.is_connected(),
    }


@router.get("/connections/")
def connections_check():
    return {"connections": list(ws.manager.active_connections.keys())}


@router.get("/instruments/{instrument_name}/machine")
def machine_info_by_instrument(
    instrument_name: MurfeyInstrumentName,
):
    return get_machine_config(instrument_name)[instrument_name]


@router.get("/instruments/{instrument_name}/smartem")
def check_smartem_availability(instrument_name: str):
    machine_config = get_machine_config(instrument_name)[instrument_name]
    return {"available": bool(machine_config.smartem_api_url)}


@router.get("/instruments/{instrument_name}/visits_raw", response_model=List[Visit])
def get_current_visits(instrument_name: MurfeyInstrumentName, db=ispyb_db):
    logger.debug(
        f"Received request to look up ongoing visits for {sanitise(instrument_name)}"
    )
    return get_all_ongoing_visits(instrument_name, db)


@router.get("/instruments/{instrument_name}/visits/")
def all_visit_info(
    instrument_name: MurfeyInstrumentName, request: Request, db=ispyb_db
):
    visits = get_all_ongoing_visits(instrument_name, db)

    if visits:
        return_query = [
            {
                "Start date": visit.start,
                "End date": visit.end,
                "Visit name": visit.name,
                "Time remaining": str(visit.end - datetime.now()),
            }
            for visit in visits
        ]  # "Proposal title": visit.proposal_title
        logger.debug(
            f"{len(visits)} visits active for {sanitise(instrument_name)=}: {', '.join(v.name for v in visits)}"
        )
        return templates.TemplateResponse(
            request=request,
            name="activevisits.html",
            context={"info": return_query, "microscope": instrument_name},
        )
    else:
        logger.debug(f"No visits identified for {sanitise(instrument_name)=}")
        return templates.TemplateResponse(
            request=request,
            name="activevisits.html",
            context={"info": [], "microscope": instrument_name},
        )


@router.get("/sessions/{session_id}/rsyncers", response_model=List[RsyncInstance])
def get_rsyncers_for_client(session_id: MurfeySessionID, db=murfey_db):
    rsync_instances = db.exec(
        select(RsyncInstance).where(RsyncInstance.session_id == session_id)
    )
    return rsync_instances.all()


class SessionClients(BaseModel):
    session: Session
    clients: List[ClientEnvironment]


@router.get("/sessions/{session_id}")
async def get_session(session_id: MurfeySessionID, db=murfey_db) -> SessionClients:
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    clients = db.exec(
        select(ClientEnvironment).where(ClientEnvironment.session_id == session_id)
    ).all()
    return SessionClients(session=session, clients=clients)


@router.get("/sessions")
async def get_sessions(db=murfey_db):
    sessions = db.exec(select(Session)).all()
    clients = db.exec(select(ClientEnvironment)).all()
    res = []
    for sess in sessions:
        r = {"session": sess, "clients": []}
        for cl in clients:
            if cl.session_id == sess.id:
                r["clients"].append(cl)
        res.append(r)
    return res


class VisitEndTime(BaseModel):
    end_time: Optional[datetime] = None


@router.post("/instruments/{instrument_name}/visits/{visit}/sessions/{name}")
def create_session(
    instrument_name: MurfeyInstrumentName,
    visit: str,
    name: str,
    visit_end_time: VisitEndTime,
    db=murfey_db,
) -> int:
    s = Session(
        name=name,
        visit=visit,
        instrument_name=instrument_name,
        visit_end_time=visit_end_time.end_time,
    )
    db.add(s)
    db.commit()
    sid = s.id
    return sid


@router.post("/sessions/{session_id}")
def update_session(
    session_id: MurfeySessionID, process: bool = True, db=murfey_db
) -> None:
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    session.process = process
    db.add(session)
    db.commit()
    return None


@router.delete("/sessions/{session_id}")
def remove_session(session_id: MurfeySessionID, db=murfey_db):
    remove_session_by_id(session_id, db)


@router.get("/instruments/{instrument_name}/visits/{visit_name}/sessions")
def get_sessions_with_visit(
    instrument_name: MurfeyInstrumentName, visit_name: str, db=murfey_db
) -> List[Session]:
    sessions = db.exec(
        select(Session)
        .where(Session.instrument_name == instrument_name)
        .where(Session.visit == visit_name)
    ).all()
    return sessions


@router.get("/instruments/{instrument_name}/sessions")
async def get_sessions_by_instrument_name(
    instrument_name: MurfeyInstrumentName, db=murfey_db
) -> List[Session]:
    sessions = db.exec(
        select(Session).where(Session.instrument_name == instrument_name)
    ).all()
    return sessions


@router.get("/sessions/{session_id}/data_collection_groups")
def get_dc_groups(
    session_id: MurfeySessionID, db=murfey_db
) -> Dict[str, DataCollectionGroup]:
    data_collection_groups = db.exec(
        select(DataCollectionGroup).where(DataCollectionGroup.session_id == session_id)
    ).all()
    return {dcg.tag: dcg for dcg in data_collection_groups}


@router.get("/sessions/{session_id}/data_collection_groups/{dcgid}/data_collections")
def get_data_collections(
    session_id: MurfeySessionID, dcgid: int, db=murfey_db
) -> List[DataCollection]:
    data_collections = db.exec(
        select(DataCollection).where(DataCollection.dcg_id == dcgid)
    ).all()
    return data_collections


@router.get("/clients")
async def get_clients(db=murfey_db):
    clients = db.exec(select(ClientEnvironment)).all()
    return clients


class CurrentGainRef(BaseModel):
    path: str


@router.put("/sessions/{session_id}/current_gain_ref")
def update_current_gain_ref(
    session_id: MurfeySessionID, new_gain_ref: CurrentGainRef, db=murfey_db
):
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    session.current_gain_ref = new_gain_ref.path
    db.add(session)

    session_processing_parameters = db.exec(
        select(SessionProcessingParameters).where(
            SessionProcessingParameters.session_id == session_id
        )
    ).all()
    if session_processing_parameters:
        session_processing_parameters[0].gain_ref = new_gain_ref.path
        db.add(session_processing_parameters[0])

    db.commit()


spa_router = APIRouter(
    prefix="/session_info/spa",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: SPA"],
)


class ProcessingDetails(BaseModel):
    data_collection_group: DataCollectionGroup
    data_collections: List[DataCollection]
    processing_jobs: List[ProcessingJob]
    relion_params: SPARelionParameters
    feedback_params: ClassificationFeedbackParameters


@spa_router.get("/sessions/{session_id}/spa_processing_parameters")
def get_spa_proc_param_details(
    session_id: MurfeySessionID, db=murfey_db
) -> Optional[List[ProcessingDetails]]:
    params = db.exec(
        select(
            DataCollectionGroup,
            DataCollection,
            ProcessingJob,
            SPARelionParameters,
            ClassificationFeedbackParameters,
        )
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollectionGroup.id == DataCollection.dcg_id)
        .where(DataCollection.id == ProcessingJob.dc_id)
        .where(SPARelionParameters.pj_id == ProcessingJob.id)
        .where(ClassificationFeedbackParameters.pj_id == ProcessingJob.id)
    ).all()
    if not params:
        return None
    unique_dcg_indices = []
    dcg_ids = []
    for i, p in enumerate(params):
        if p[0].id not in dcg_ids:
            dcg_ids.append(p[0].id)
            unique_dcg_indices.append(i)

    def _parse(ps, i, dcg_id):
        res = []
        for p in ps:
            if p[0].id == dcg_id:
                if p[i] not in res:
                    res.append(p[i])
        return res

    return [
        ProcessingDetails(
            data_collection_group=params[i][0],
            data_collections=_parse(params, 1, d),
            processing_jobs=_parse(params, 2, d),
            relion_params=_parse(params, 3, d)[0],
            feedback_params=_parse(params, 4, d)[0],
        )
        for i, d in zip(unique_dcg_indices, dcg_ids)
    ]


@spa_router.get(
    "/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares/{gsid}/foil_holes/{fhid}/num_movies"
)
def get_number_of_movies_from_foil_hole(
    session_id: int, dcgid: int, gsid: int, fhid: int, db=murfey_db
) -> int:
    movies = db.exec(
        select(Movie, FoilHole, GridSquare, DataCollectionGroup)
        .where(Movie.foil_hole_id == FoilHole.id)
        .where(FoilHole.name == fhid)
        .where(FoilHole.grid_square_id == GridSquare.id)
        .where(GridSquare.name == gsid)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return len(movies)


@spa_router.get("/sessions/{session_id}/grid_squares")
def get_grid_squares(session_id: MurfeySessionID, db=murfey_db):
    return _get_grid_squares(session_id, db)


@spa_router.get("/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares")
def get_grid_squares_from_dcg(
    session_id: MurfeySessionID, dcgid: int, db=murfey_db
) -> List[GridSquare]:
    return _get_grid_squares_from_dcg(session_id, dcgid, db)


@spa_router.get(
    "/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares/{gsid}/foil_holes"
)
def get_foil_holes_from_grid_square(
    session_id: MurfeySessionID, dcgid: int, gsid: int, db=murfey_db
) -> List[FoilHole]:
    return _get_foil_holes_from_grid_square(session_id, dcgid, gsid, db)


@spa_router.get("/sessions/{session_id}/foil_hole/{fh_name}")
def get_foil_hole(
    session_id: MurfeySessionID, fh_name: int, db=murfey_db
) -> Dict[str, int]:
    return _get_foil_hole(session_id, fh_name, db)


tomo_router = APIRouter(
    prefix="/session_info/tomo",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: CryoET"],
)


@tomo_router.get("/sessions/{session_id}/tilt_series/{tilt_series_tag}/tilts")
def get_tilts(
    session_id: MurfeySessionID, tilt_series_tag: str, db=murfey_db
) -> Dict[str, List[str]]:
    res = db.exec(
        select(TiltSeries, Tilt)
        .where(TiltSeries.tag == tilt_series_tag)
        .where(TiltSeries.session_id == session_id)
        .where(Tilt.tilt_series_id == TiltSeries.id)
    ).all()
    tilts: Dict[str, List[str]] = {}
    for el in res:
        if tilts.get(el[1].rsync_source):
            tilts[el[1].rsync_source].append(el[2].movie_path)
        else:
            tilts[el[1].rsync_source] = [el[2].movie_path]
    return tilts


correlative_router = APIRouter(
    prefix="/session_info/correlative",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: Correlative Imaging"],
)


@correlative_router.get("/sessions/{session_id}/upstream_visits")
async def find_upstream_visits(session_id: MurfeySessionID, db=murfey_db):
    return _find_upstream_visits(session_id=session_id, db=db)


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_file_paths"
)
async def gather_upstream_files(
    visit_name: str,
    session_id: MurfeySessionID,
    upstream_file_request: UpstreamFileRequestInfo,
    db=murfey_db,
):
    return _gather_upstream_files(
        session_id=session_id,
        upstream_instrument=upstream_file_request.upstream_instrument,
        upstream_visit_path=upstream_file_request.upstream_visit_path,
        db=db,
    )


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_file/{upstream_file_path:path}"
)
async def get_upstream_file(
    visit_name: str,
    session_id: MurfeySessionID,
    upstream_file_path: Path,
    db=murfey_db,
):
    upstream_file = _get_upstream_file(upstream_file_path)
    return (
        FileResponse(path=upstream_file) if upstream_file is not None else upstream_file
    )


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_tiff_paths"
)
async def gather_upstream_tiffs(visit_name: str, session_id: int, db=murfey_db):
    return _gather_upstream_tiffs(visit_name=visit_name, session_id=session_id, db=db)


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_tiff/{tiff_path:path}"
)
async def get_tiff_file(visit_name: str, session_id: int, tiff_path: str, db=murfey_db):
    tiff_file = _get_tiff_file(
        visit_name=visit_name, session_id=session_id, tiff_path=tiff_path, db=db
    )
    return FileResponse(path=tiff_file) if isinstance(tiff_file, Path) else tiff_file



import queue
import threading
from datetime import datetime
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from murfey.client.rsync import RSyncer
from tests.conftest import ExampleVisit


@pytest.fixture
def rsync_module():
    return "data"


@pytest.fixture
def mock_server_url():
    mock_url = MagicMock()
    mock_url.hostname = "10.0.0.1"
    return mock_url


# Create a dummy callback function
def dummy_callback():
    return None


@pytest.mark.parametrize("is_local", (True, False))
def test_rsyncer_initialises(
    tmp_path: Path,
    rsync_module: str,
    mock_server_url: MagicMock,
    is_local: bool,
):
    # Assign values to parameters
    basepath_local = tmp_path / "local"
    basepath_remote = tmp_path / "remote"

    # Create a test substrings blacklist dict
    substrings_blacklist = {
        "directories": ["1", "2", "3"],
        "files": ["a", "b", "c"],
    }

    # Create a timestamp
    timestamp = datetime.now()

    # Initialise the RSyncer
    rsyncer = RSyncer(
        basepath_local=basepath_local,
        basepath_remote=basepath_remote,
        rsync_module=rsync_module,
        server_url=mock_server_url,
        stop_callback=dummy_callback,
        local=is_local,
        substrings_blacklist=substrings_blacklist,
        end_time=timestamp,
    )

    # Check that the attributes are as expected
    assert rsyncer._basepath == basepath_local.absolute()
    assert rsyncer._basepath_remote == basepath_remote
    assert rsyncer._rsync_module == rsync_module
    assert rsyncer._server_url == mock_server_url
    assert rsyncer._stop_callback == dummy_callback
    assert rsyncer._local == is_local
    assert rsyncer._do_transfer
    assert not rsyncer._remove_files
    assert rsyncer._required_substrings_for_removal == []
    assert rsyncer._substrings_blacklist == substrings_blacklist
    assert rsyncer._notify
    assert rsyncer._end_time == timestamp
    assert rsyncer._skipped_files == []
    assert (
        rsyncer._remote == str(basepath_remote)
        if is_local
        else f"{mock_server_url.hostname}::{rsync_module}/{basepath_remote}"
    )
    assert rsyncer._files_transferred == 0
    assert rsyncer._bytes_transferred == 0
    assert isinstance(rsyncer.queue, queue.Queue)
    assert isinstance(rsyncer.thread, threading.Thread)
    assert not rsyncer._stopping
    assert not rsyncer._halt_thread
    assert not rsyncer._finalising
    assert not rsyncer._finalised


@pytest.mark.parametrize(
    "test_params",
    (
        # Is stopping? | Is thread alive? | Expected status
        (False, False, "ready"),
        (False, True, "running"),
        (True, True, "stopping"),
        (True, False, "finished"),
    ),
)
def test_rsyncer_status(
    tmp_path: Path,
    mock_server_url: MagicMock,
    test_params: tuple[bool, bool, str],
):
    # Unpack test params
    is_stopping, is_thread_alive, expected_status = test_params

    # Mock the thread
    mock_thread = MagicMock()
    mock_thread.is_alive.return_value = is_thread_alive

    # Initialise the RSyncer and patch the attributes to be tested
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer.thread = mock_thread
    rsyncer._stopping = is_stopping

    # Check that its status is correct
    assert rsyncer.status == expected_status

    # Check that its canonical representation is correct
    assert str(rsyncer) == f"<RSyncer ({rsyncer._basepath} → {rsyncer._remote})>"


@pytest.mark.parametrize("notify", (True, False))
def test_rsyncer_notify(
    mocker: MockerFixture,
    tmp_path: Path,
    mock_server_url: MagicMock,
    notify: bool,
):
    # Patch the superclass that RSyncer stems from
    mock_notify = mocker.patch("murfey.client.rsync.Observer.notify")
    mock_notify.return_value = None

    # Initialise the RSyncer
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
        notify=notify,
    )
    # Check that the 'notify' attribute is set correctly
    assert rsyncer._notify == notify

    # Run 'notify' and check that the expected calls were made
    rsyncer.notify("arg1", "arg2", kwarg1="kwarg1", kwarg2="kwarg2")
    if notify:
        mock_notify.assert_called_once_with(
            "arg1",
            "arg2",
            secondary=False,
            kwarg1="kwarg1",
            kwarg2="kwarg2",
        )
    else:
        mock_notify.assert_not_called()


@pytest.mark.parametrize("rsyncer_status", ("default", "is_alive", "stopping"))
def test_rsyncer_start(
    tmp_path: Path,
    mock_server_url: MagicMock,
    rsyncer_status: str,
):
    # Mock the thread attribute so that it doesn't start an actual Thread
    mock_thread = MagicMock()
    mock_thread.start.return_value = None
    mock_thread.is_alive.return_value = rsyncer_status == "is_alive"

    # Initialise the RSyncer and patch the attributes to be tested
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer.thread = mock_thread
    rsyncer._stopping = rsyncer_status == "stopping"

    # Start the RSyncer
    if rsyncer_status == "default":
        rsyncer.start()
        mock_thread.start.assert_called_once()
    else:
        with pytest.raises(RuntimeError):
            rsyncer.start()


def test_rsyncer_restart(
    mocker: MockerFixture,
    tmp_path: Path,
    mock_server_url: MagicMock,
):
    # Patch the 'start' class method, which is called by 'restart'
    mock_start = mocker.patch.object(RSyncer, "start")
    mock_start.return_value = None

    # Mock the thread and the attributes used
    mock_thread = MagicMock()
    mock_thread.join.return_value = None

    # Initialise the RSyncer and patch the attributes to be tested
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer.thread = mock_thread

    # Run 'restart'
    rsyncer.restart()

    # Check that the correct calls and attributes are present
    mock_thread.join.assert_called_once()
    assert not rsyncer._halt_thread
    assert isinstance(rsyncer.thread, threading.Thread)
    mock_start.assert_called_once()


@pytest.mark.parametrize("thread_is_alive", (True, False))
def test_rsyncer_stop(
    tmp_path: Path,
    mock_server_url: MagicMock,
    thread_is_alive: bool,
):
    # Mock the thread
    mock_thread = MagicMock()
    mock_thread.is_alive.return_value = thread_is_alive
    mock_thread.join.return_value = None

    # Mock the queue
    mock_queue = MagicMock()
    mock_queue.join.return_value = None
    mock_queue.put.return_value = None

    # Initialise the RSyncer and patch the attributes to be tested
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer.thread = mock_thread
    rsyncer.queue = mock_queue

    # Check that initial attributes are as expected
    assert not rsyncer._stopping
    assert not rsyncer._halt_thread

    # Run 'stop' and check that the calls are as expected
    rsyncer.stop()

    assert rsyncer._stopping
    assert rsyncer._halt_thread
    if thread_is_alive:
        mock_queue.join.assert_called_once()
        mock_queue.put.assert_called_with(None)
        mock_thread.join.assert_called_once()
    else:
        mock_queue.join.assert_not_called()
        mock_queue.put.assert_not_called()
        mock_thread.join.assert_not_called()


def test_rsyncer_request_stop(
    tmp_path: Path,
    mock_server_url: MagicMock,
):
    # Initialise the RSyncer
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )

    # Check that initial attributes are as expected
    assert not rsyncer._stopping
    assert not rsyncer._halt_thread

    # Run 'request_stop' and check that attributes have changed
    rsyncer.request_stop()
    assert rsyncer._stopping
    assert rsyncer._halt_thread


@pytest.fixture
def clem_visit_dir(tmp_path: Path):
    visit_name = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}-{ExampleVisit.visit_number}"
    visit_dir = tmp_path / "local" / visit_name
    visit_dir.mkdir(parents=True, exist_ok=True)
    return visit_dir


@pytest.fixture
def clem_test_files(clem_visit_dir: Path):
    # Create test files for the DirWatcher to scan
    file_list: list[Path] = []
    project_dir = clem_visit_dir / "images" / "test_grid"

    # Example atlas collection
    for s in range(20):
        file_list.append(
            project_dir
            / "Overview 1"
            / "Image 1"
            / f"Image 1--Stage{str(s).zfill(2)}.tif"
        )
    file_list.append(
        project_dir / "Overview 1" / "Image 1" / "Metadata" / "Image 1.xlif"
    )

    # Example image stack collection
    for c in range(3):
        for z in range(10):
            file_list.append(
                project_dir
                / "TileScan 1"
                / "Position 1"
                / f"Position 1--C{str(c).zfill(2)}--Z{str(z).zfill(2)}.tif"
            )
    file_list.append(
        project_dir / "TileScan 1" / "Position 1" / "Metadata" / "Position 1.xlif"
    )

    # Create all files and directories specified
    for file in file_list:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        if not file.exists():
            file.touch()
    return sorted(file_list)


@pytest.fixture
def clem_junk_files(clem_visit_dir: Path):
    # Create junk files that are to be blacklisted from the CLEM workflow
    file_list: list[Path] = []
    project_dir = clem_visit_dir / "images" / "test_grid"

    # Create junk atlas data
    for n in range(5):
        for s in range(20):
            file_list.append(
                project_dir
                / "Image 1"
                / f"Image 1_pmd_{n}"
                / f"Image 1_pmd_{n}--Stage{str(s).zfill(2)}.tif"
            )
        file_list.append(
            project_dir / "Image 1" / f"Image 1_pmd_{n}" / "Metadata" / "Image 1.xlif"
        )

    # Creat junk image data
    for n in range(5):
        for c in range(3):
            for z in range(10):
                file_list.append(
                    project_dir
                    / "Position 1"
                    / f"Position 1_pmd_{n}"
                    / f"Position 1_pmd_{n}--C{str(c).zfill(2)}--Z{str(z).zfill(2)}.tif"
                )
        file_list.append(
            project_dir
            / "Position 1"
            / f"Position 1_pmd_{n}"
            / "Metadata"
            / "Position 1.xlif"
        )

    # Create remaining junk files
    for file_path in (
        "1.xlef",
        "Metadata/IOManagerConfiguation.xlif",
        "Metadata/Overview 1.xlcf",
        "Metadata/TileScan 1.xlcf",
        "Overview 1/Image 1/Image 1_histo.lof",
        "TileScan 1/Position 1/Position 1_histo.lof",
        "Overview 1/Image 1/Metadata/Image 1_histo.xlif",
        "TileScan 1/Position 1/Metadata/Position 1_histo.xlif",
    ):
        file_list.append(project_dir / file_path)

    # Create files and directoriees
    for file in file_list:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        if not file.exists():
            file.touch()
    return sorted(file_list)


clem_substrings_blacklist = {
    "directories": [
        "_pmd_",
    ],
    "files": [
        ".xlef",
        ".xlcf",
        "_histo.lof",
        "_histo.xlif",
        "IOManagerConfiguation.xlif",
    ],
}


@pytest.mark.parametrize(
    "test_params",
    (
        # Workflow type | Use thread? | Use callback function? | Use blacklist?
        ("clem", False, False, False),
        ("clem", False, False, True),
        ("clem", False, True, False),
        ("clem", False, True, True),
        ("clem", True, False, False),
        ("clem", True, False, True),
        ("clem", True, True, False),
        ("clem", True, True, True),
    ),
)
def test_rsyncer_finalise(
    mocker: MockerFixture,
    rsync_module: str,
    mock_server_url: MagicMock,
    clem_visit_dir: Path,
    clem_test_files: list[Path],
    clem_junk_files: list[Path],
    test_params: tuple[str, bool, bool, bool],
):
    # Unpack test params
    workflow_type, use_thread, use_callback, use_blacklist = test_params

    # Create a test end time
    timestamp = datetime.now()

    # Mock the class functions/attributes called by the 'finalise' class function
    mock_queue = MagicMock()
    mock_queue.put.return_value = None

    mock_transfer = mocker.patch.object(RSyncer, "_transfer")
    mock_transfer.return_value = True

    mock_stop = mocker.patch.object(RSyncer, "stop")
    mock_stop.return_value = None

    mock_process = mocker.patch.object(RSyncer, "_process")
    mock_process.return_value = None

    mock_callback = MagicMock(return_value=None)

    # Initialise the RSyncer class based on the workflow type being tested
    if workflow_type == "clem":
        rsyncer = RSyncer(
            basepath_local=clem_visit_dir / "images",
            basepath_remote=Path(clem_visit_dir.name) / "images",
            rsync_module=rsync_module,
            server_url=mock_server_url,
            stop_callback=dummy_callback,
            substrings_blacklist=clem_substrings_blacklist if use_blacklist else {},
            end_time=timestamp,
        )
        # Patch the 'queue' attribute with the mocked one
        rsyncer.queue = mock_queue

        # Check the initial state of attributes that will be changed by 'finalise'
        assert not rsyncer._remove_files
        assert rsyncer._notify
        assert rsyncer._end_time == timestamp
        assert not rsyncer._finalising
        assert not rsyncer._finalised

        # Run the 'finalise' class function with the workflow-specific paths
        rsyncer.finalise(
            thread=use_thread,
            callback=mock_callback if use_callback else None,
        )

        # Check that attributes are set correctly at the start of the function
        assert rsyncer._remove_files
        assert not rsyncer._notify
        assert rsyncer._end_time is None
        assert rsyncer._finalising

        # Check that list of files with and without using a blacklist are correct
        if use_thread:
            for file in clem_test_files:
                mock_queue.put.assert_any_call(file)
            if not use_blacklist:
                for file in clem_junk_files:
                    mock_queue.put.assert_any_call(file)
        else:
            transfer_args = mock_transfer.call_args.args
            assert sorted(transfer_args[0]) == (
                sorted(clem_test_files)
                if use_blacklist
                else sorted([*clem_test_files, *clem_junk_files])
            )

        # Transfer is being mocked, so check that files to transfer are all present
        for file in clem_test_files:
            assert file.exists()
        for file in clem_junk_files:
            assert not file.exists() if use_blacklist else file.exists()

        # Check that stop was called the correct number of times depending on the setup
        assert mock_stop.call_count == 2 if use_thread else 1

        # Check that the RSyncer is marked as finalised at the end
        assert rsyncer._finalised

        # Check that the callback was set at the end
        if use_callback:
            mock_callback.assert_called_once()


@pytest.mark.parametrize("is_stopping", (True, False))
def test_rsyncer_enqueue(
    tmp_path: Path,
    mock_server_url: MagicMock,
    is_stopping: bool,
):
    # Mock the queue
    mock_queue = MagicMock()
    mock_queue.put.return_value = None

    # Initialise the RSyncer and patch the attributes used by the test
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer._stopping = is_stopping
    rsyncer.queue = mock_queue

    # Run enqueue with a test file and check that the expected calls were made
    rsyncer.enqueue(Path("test_file"))
    if is_stopping:
        mock_queue.put.assert_not_called()
    else:
        mock_queue.put.assert_called_once_with(
            (tmp_path / "local" / "test_file").absolute()
        )


def test_rsyncer_flush_skipped(
    tmp_path: Path,
    mock_server_url: MagicMock,
):
    # Mock the queue
    mock_queue = MagicMock()
    mock_queue.put.return_value = None

    # Create a list of test files
    skipped_files = [
        tmp_path / "local" / f"file_{str(n).zfill(2)}.txt" for n in range(20)
    ]

    # Initialise the RSyncer and patch the attributes used by the test
    rsyncer = RSyncer(
        basepath_local=tmp_path / "local",
        basepath_remote=tmp_path / "remote",
        rsync_module=mock.ANY,
        server_url=mock_server_url,
    )
    rsyncer.queue = mock_queue
    rsyncer._skipped_files = skipped_files

    # Run 'flush_skipped' and check that it works as intended
    rsyncer.flush_skipped()
    for f in skipped_files:
        mock_queue.put.assert_any_call(f)
    assert rsyncer._skipped_files == []



import logging
from datetime import datetime

import ispyb.sqlalchemy._auto_db_schema as ISPyBDB
from sqlmodel import select
from sqlmodel.orm.session import Session as SQLModelSession

import murfey.server.prometheus as prom
import murfey.util.db as MurfeyDB
from murfey.server import _transport_object
from murfey.server.ispyb import ISPyBSession
from murfey.util import sanitise

logger = logging.getLogger("murfey.workflows.register_processing_job")


def run(message: dict, murfey_db: SQLModelSession):
    # Faill immediately if not transport manager is set
    if _transport_object is None:
        logger.error("Unable to find transport manager")
        return {"success": False, "requeue": False}

    logger.info(f"Registering the following processing job: \n{message}")

    murfey_session_id = message["session_id"]
    dc = murfey_db.exec(
        select(MurfeyDB.DataCollection, MurfeyDB.DataCollectionGroup)
        .where(MurfeyDB.DataCollection.dcg_id == MurfeyDB.DataCollectionGroup.id)
        .where(MurfeyDB.DataCollectionGroup.session_id == murfey_session_id)
        .where(MurfeyDB.DataCollectionGroup.tag == message["source"])
        .where(MurfeyDB.DataCollection.tag == message["tag"])
    ).all()

    if dc:
        _dcid = dc[0][0].id
    else:
        logger.warning(f"No data collection ID found for {sanitise(message['tag'])}")
        return {"success": False, "requeue": True}
    if pj_murfey := murfey_db.exec(
        select(MurfeyDB.ProcessingJob)
        .where(MurfeyDB.ProcessingJob.recipe == message["recipe"])
        .where(MurfeyDB.ProcessingJob.dc_id == _dcid)
    ).all():
        pid = pj_murfey[0].id
    else:
        if ISPyBSession() is None:
            murfey_pj = MurfeyDB.ProcessingJob(recipe=message["recipe"], dc_id=_dcid)
        else:
            record = ISPyBDB.ProcessingJob(
                dataCollectionId=_dcid, recipe=message["recipe"]
            )
            run_parameters = message.get("parameters", {})
            assert isinstance(run_parameters, dict)
            if message.get("job_parameters"):
                job_parameters = [
                    ISPyBDB.ProcessingJobParameter(parameterKey=k, parameterValue=v)
                    for k, v in message["job_parameters"].items()
                ]
                pid = _transport_object.do_create_ispyb_job(
                    record, params=job_parameters
                ).get("return_value", None)
            else:
                pid = _transport_object.do_create_ispyb_job(record).get(
                    "return_value", None
                )
            if pid is None:
                return {"success": False, "requeue": True}
            murfey_pj = MurfeyDB.ProcessingJob(
                id=pid, recipe=message["recipe"], dc_id=_dcid
            )
        murfey_db.add(murfey_pj)
        murfey_db.commit()
        pid = murfey_pj.id
        murfey_db.close()

    # Update Prometheus counter for preprocessed movies
    prom.preprocessed_movies.labels(processing_job=pid)

    # Register AutoProcProgram database entry if it doesn't already exist
    if not murfey_db.exec(
        select(MurfeyDB.AutoProcProgram).where(MurfeyDB.AutoProcProgram.pj_id == pid)
    ).all():
        if ISPyBSession() is None:
            murfey_app = MurfeyDB.AutoProcProgram(pj_id=pid)
        else:
            record = ISPyBDB.AutoProcProgram(
                processingJobId=pid, processingStartTime=datetime.now()
            )
            appid = _transport_object.do_update_processing_status(record).get(
                "return_value", None
            )
            if appid is None:
                return {"success": False, "requeue": True}
            murfey_app = MurfeyDB.AutoProcProgram(id=appid, pj_id=pid)
        murfey_db.add(murfey_app)
        murfey_db.commit()
        murfey_db.close()
    return {"success": True}



from __future__ import annotations

import logging
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, List, NamedTuple

import xmltodict

from murfey.client.instance_environment import MurfeyInstanceEnvironment, SampleInfo
from murfey.util.client import capture_post, get_machine_config_client

logger = logging.getLogger("murfey.client.context")


def _atlas_destination(
    environment: MurfeyInstanceEnvironment, source: Path, token: str
) -> Path:
    machine_config = get_machine_config_client(
        str(environment.url.geturl()),
        token,
        instrument_name=environment.instrument_name,
    )
    for i, destination_part in enumerate(
        Path(environment.default_destinations[source]).parts
    ):
        if destination_part == environment.visit:
            return Path(machine_config.get("rsync_basepath", "")) / "/".join(
                Path(environment.default_destinations[source]).parent.parts[: i + 1]
            )
    return (
        Path(machine_config.get("rsync_basepath", ""))
        / Path(environment.default_destinations[source]).parent
        / environment.visit
    )


def ensure_dcg_exists(
    collection_type: str,
    metadata_source: Path,
    environment: MurfeyInstanceEnvironment,
    token: str,
) -> str | None:
    """Create  a data collection group"""
    if collection_type == "tomo":
        experiment_type_id = 36
        session_file = metadata_source / "Session.dm"
        source_visit_dir = metadata_source.parent
    elif collection_type == "spa":
        experiment_type_id = 37
        # For SPA the metadata source sent should include the Images-Disc
        session_file = metadata_source.parent / "EpuSession.dm"
        source_visit_dir = metadata_source.parent.parent
        for h in entry_points(group="murfey.hooks"):
            try:
                if h.name == "get_epu_session_metadata":
                    h.load()(
                        destination_dir=session_file.parent,
                        environment=environment,
                        token=token,
                    )
            except Exception as e:
                logger.warning(f"Get EPU session hook failed: {e}")
    else:
        logger.error(f"Unknown collection type {collection_type}")
        return None

    if not session_file.is_file():
        logger.warning(f"Cannot find session file {str(session_file)}")
        dcg_tag = "/".join(
            [part for part in metadata_source.parts if part != environment.visit]
        ).replace("//", "/")
        dcg_data = {
            "experiment_type_id": experiment_type_id,
            "tag": dcg_tag,
        }
    else:
        with open(session_file, "r") as session_xml:
            session_data = xmltodict.parse(session_xml.read())

        if collection_type == "tomo":
            windows_path = session_data["TomographySession"]["AtlasId"]
        else:
            windows_path = session_data["EpuSessionXml"]["Samples"]["_items"][
                "SampleXml"
            ][0]["AtlasId"]["#text"]

        logger.info(f"Windows path to atlas metadata found: {windows_path}")
        if not windows_path:
            logger.warning("No atlas metadata path found")
            return None
        if environment.visit in windows_path.split("\\"):
            # Case of atlas in the correct location
            visit_index = windows_path.split("\\").index(environment.visit)
            partial_path = "/".join(windows_path.split("\\")[visit_index + 1 :])
            logger.info(
                f"Partial Linux path successfully constructed from Windows path: {partial_path}"
            )
        else:
            # Atlas not in visit, so come up with where it should have been
            # Assumes /structure/to/Supervisor/Sample/Atlas/Atlas.dm
            partial_path = "atlas/" + "/".join(windows_path.split("\\")[-4:])
            logger.info(f"Partial Linux path estimated: {partial_path}")

        logger.info(
            f"Looking for atlas XML file in metadata directory {str((source_visit_dir / partial_path).parent)}"
        )

        dcg_tag = "/".join(
            [part for part in metadata_source.parts if part != environment.visit]
        ).replace("//", "/")

        for p in partial_path.split("/"):
            if p.startswith("Sample"):
                sample = int(p.replace("Sample", ""))
                break
        else:
            logger.warning(f"Sample could not be identified for {metadata_source}")
            return None
        environment.samples[metadata_source] = SampleInfo(
            atlas=Path(partial_path), sample=sample
        )

        if atlas_xml_search := list(
            (source_visit_dir / partial_path).parent.glob("Atlas_*.xml")
        ):
            atlas_xml_path = atlas_xml_search[0]
            logger.info(f"Atlas XML path {str(atlas_xml_path)} found")
            with open(atlas_xml_path, "rb") as atlas_xml:
                atlas_xml_data = xmltodict.parse(atlas_xml)
                atlas_original_pixel_size = float(
                    atlas_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
                        "numericValue"
                    ]
                )
            # need to calculate the pixel size of the downscaled image
            atlas_pixel_size = atlas_original_pixel_size * 7.8
            logger.info(f"Atlas image pixel size determined to be {atlas_pixel_size}")

            dcg_data = {
                "experiment_type_id": experiment_type_id,
                "tag": dcg_tag,
                "atlas": str(
                    _atlas_destination(environment, session_file.parent, token)
                    / environment.samples[metadata_source].atlas.parent
                    / atlas_xml_path.with_suffix(".jpg").name
                ).replace("//", "/"),
                "sample": environment.samples[metadata_source].sample,
                "atlas_pixel_size": atlas_pixel_size,
            }
        else:
            dcg_data = {
                "experiment_type_id": experiment_type_id,
                "tag": dcg_tag,
                "sample": environment.samples[metadata_source].sample,
            }
    capture_post(
        base_url=str(environment.url.geturl()),
        router_name="workflow.router",
        function_name="register_dc_group",
        token=token,
        visit_name=environment.visit,
        session_id=environment.murfey_session,
        data=dcg_data,
    )
    return dcg_tag


class ProcessingParameter(NamedTuple):
    name: str
    label: str
    default: Any = None


def detect_acquisition_software(dir_for_transfer: Path) -> str:
    glob = dir_for_transfer.glob("*")
    for f in glob:
        if f.name.startswith("EPU") or f.name.startswith("GridSquare"):
            return "epu"
        if f.name.startswith("Position") or f.suffix == ".mdoc":
            return "tomo"
    return ""


class Context:
    user_params: List[ProcessingParameter] = []
    metadata_params: List[ProcessingParameter] = []

    def __init__(self, name: str, acquisition_software: str, token: str):
        self._acquisition_software = acquisition_software
        self._token = token
        self.name = name
        self.data_collection_parameters: dict = {}

    def post_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        # Search external packages for additional hooks to include in Murfey
        for h in entry_points(group="murfey.post_transfer_hooks"):
            if h.name == self.name:
                h.load()(transferred_file, environment=environment, **kwargs)

    def post_first_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        self.post_transfer(transferred_file, environment=environment, **kwargs)

    def gather_metadata(
        self, metadata_file: Path, environment: MurfeyInstanceEnvironment | None = None
    ):
        raise NotImplementedError(
            f"gather_metadata must be declared in derived class to be used: {self}"
        )



from __future__ import annotations

import json
import logging
import threading
import time
from queue import Empty, Queue

import requests


class LogFilter(logging.Filter):
    """A filter to limit messages going to Graylog"""

    def __repr__(self):
        return "<murfey.server.LogFilter>"

    def __init__(self):
        super().__init__()
        self._filter_levels = {
            "murfey": logging.DEBUG,
            "ispyb": logging.DEBUG,
            "zocalo": logging.DEBUG,
            "uvicorn": logging.INFO,
            "fastapi": logging.INFO,
            "starlette": logging.INFO,
            "sqlalchemy": logging.INFO,
        }

    @staticmethod
    def install() -> LogFilter:
        logfilter = LogFilter()
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            handler.addFilter(logfilter)
        return logfilter

    def filter(self, record: logging.LogRecord) -> bool:
        logger_name = record.name
        while True:
            if logger_name in self._filter_levels:
                return record.levelno >= self._filter_levels[logger_name]
            if "." not in logger_name:
                return False
            logger_name = logger_name.rsplit(".", maxsplit=1)[0]


class HTTPSHandler(logging.Handler):
    """
    A log handler collects log messages and posts them in batches to the backend
    FastAPI server using HTTPS POST.
    """

    def __init__(
        self,
        endpoint_url: str,
        min_batch: int = 5,
        max_batch: int = 50,
        min_interval: float = 0.5,
        max_interval: float = 2.0,
        max_retry: int = 5,
        timeout: int = 3,
        token: str = "",
    ):
        super().__init__()
        self.endpoint_url = endpoint_url
        self.queue: Queue = Queue()
        self._stop_event = threading.Event()
        self.min_batch = min_batch
        self.max_batch = max_batch
        self.min_interval = min_interval
        self.max_interval = max_interval
        self.max_retry = max_retry
        self.timeout = timeout
        self.token = token

        self.log_times: list[
            float
        ] = []  # Timestamps of recent logs for rate estimation
        self.thread = threading.Thread(target=self._worker, daemon=True)
        self.thread.start()

    def emit(self, record: logging.LogRecord):
        """
        Formats the log and puts it on a queue for submission to the backend server
        """
        try:
            log_entry = self.format_record(record)
            self.queue.put(log_entry)
            self.log_times.append(time.time())
        except Exception:
            self.handleError(record)

    def format_record(self, record: logging.LogRecord):
        """
        Packages the log record as a JSON-formatted string
        """
        self.format(record)
        log_data = record.__dict__.copy()
        log_data["type"] = "log"
        return json.dumps(log_data)

    def _worker(self):
        """
        The worker function when the handler is run as a thread.
        """

        batch: list[str] = []
        last_flush = time.time()

        while not self._stop_event.is_set():
            try:
                log_entry = self.queue.get(timeout=0.05)
                batch.append(log_entry)
            # If the queue is empty, check back again
            except Empty:
                time.sleep(1)
                continue

            # Calculate logging rate based on past second
            now = time.time()
            self.log_times = [t for t in self.log_times if now - t <= 1.0]
            log_rate = len(self.log_times)

            # Adjust batch size and flush interval
            batch_size = min(max(self.min_batch, log_rate), self.max_batch)
            flush_interval = max(
                self.min_interval, min(self.max_interval, 1 / max(log_rate, 1))
            )

            # Flush if batch is ready
            if batch and (
                len(batch) >= batch_size or now - last_flush >= flush_interval
            ):
                self._send_batch(batch)
                batch = []
                last_flush = now

        # Flush remaining logs on shutdown
        if batch:
            self._send_batch(batch)

    def _send_batch(self, batch: list[str]):
        for attempt in range(0, self.max_retry):
            try:
                response = requests.post(self.endpoint_url, json=batch, timeout=5)
                if response.status_code == 200:
                    return
            except requests.RequestException:
                time.sleep(2 ** (attempt + 1) * 0.1)  # Exponential backoff

    def close(self):
        self._stop_event.set()
        self.thread.join()
        super().close()



from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional

from fastapi import APIRouter, Depends
from fastapi.responses import FileResponse
from ispyb.sqlalchemy import AutoProcProgram as ISPyBAutoProcProgram
from pydantic import BaseModel
from sqlalchemy import func
from sqlmodel import select

try:
    from smartem_backend.api_client import SmartEMAPIClient
    from smartem_common.schemas import AtlasData

    SMARTEM_ACTIVE = True
except ImportError:
    SMARTEM_ACTIVE = False

import murfey.server.prometheus as prom
from murfey.server import _transport_object
from murfey.server.api.auth import (
    MurfeySessionIDInstrument as MurfeySessionID,
    validate_instrument_token,
)
from murfey.server.api.session_shared import (
    find_upstream_visits as _find_upstream_visits,
    gather_upstream_files as _gather_upstream_files,
    gather_upstream_tiffs as _gather_upstream_tiffs,
    get_foil_hole as _get_foil_hole,
    get_foil_holes_from_grid_square as _get_foil_holes_from_grid_square,
    get_grid_squares as _get_grid_squares,
    get_grid_squares_from_dcg as _get_grid_squares_from_dcg,
    get_tiff_file as _get_tiff_file,
    get_upstream_file as _get_upstream_file,
    remove_session_by_id,
)
from murfey.server.ispyb import DB as ispyb_db, get_all_ongoing_visits
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise
from murfey.util.config import get_machine_config
from murfey.util.db import (
    AutoProcProgram,
    ClientEnvironment,
    DataCollection,
    DataCollectionGroup,
    FoilHole,
    GridSquare,
    Movie,
    ProcessingJob,
    RsyncInstance,
    Session,
)
from murfey.util.models import (
    BatchPositionParameters,
    ClientInfo,
    FoilHoleParameters,
    GridSquareParameters,
    RsyncerInfo,
    SearchMapParameters,
    UpstreamFileRequestInfo,
    Visit,
)
from murfey.workflows.spa.atlas import atlas_jpg_from_mrc
from murfey.workflows.spa.flush_spa_preprocess import (
    register_foil_hole as _register_foil_hole,
    register_grid_square as _register_grid_square,
)
from murfey.workflows.tomo.tomo_metadata import (
    register_batch_position_in_database,
    register_search_map_in_database,
)

logger = getLogger("murfey.server.api.session_control")

router = APIRouter(
    prefix="/session_control",
    dependencies=[Depends(validate_instrument_token)],
    tags=["Session Control: General"],
)


@router.get("/time")
async def get_current_timestamp():
    return {"timestamp": datetime.now().timestamp()}


@router.get("/instruments/{instrument_name}/machine")
def machine_info_by_instrument(instrument_name: str):
    return get_machine_config(instrument_name)[instrument_name]


@router.get("/new_client_id/")
async def new_client_id(db=murfey_db):
    clients = db.exec(select(ClientEnvironment)).all()
    if not clients:
        return {"new_id": 0}
    sorted_ids = sorted([c.client_id for c in clients])
    return {"new_id": sorted_ids[-1] + 1}


@router.get("/instruments/{instrument_name}/visits_raw", response_model=List[Visit])
def get_current_visits(instrument_name: str, db=ispyb_db):
    logger.debug(
        f"Received request to look up ongoing visits for {sanitise(instrument_name)}"
    )
    return get_all_ongoing_visits(instrument_name, db)


class SessionInfo(BaseModel):
    session_id: Optional[int] = None
    session_name: str = ""
    rescale: bool = True


@router.post("/instruments/{instrument_name}/clients/{client_id}/session")
def link_client_to_session(
    instrument_name: str, client_id: int, sess: SessionInfo, db=murfey_db
):
    sid = sess.session_id
    if sid is None:
        s = Session(name=sess.session_name, instrument_name=instrument_name)
        db.add(s)
        db.commit()
        sid = s.id
    client = db.exec(
        select(ClientEnvironment).where(ClientEnvironment.client_id == client_id)
    ).one()
    client.session_id = sid
    db.add(client)
    db.commit()
    db.close()
    return sid


@router.post("/visits/{visit_name}")
def register_client_to_visit(visit_name: str, client_info: ClientInfo, db=murfey_db):
    client_env = db.exec(
        select(ClientEnvironment).where(ClientEnvironment.client_id == client_info.id)
    ).one()
    session = db.exec(select(Session).where(Session.id == client_env.session_id)).one()
    if client_env:
        client_env.visit = visit_name
        db.add(client_env)
        db.commit()
    if session:
        session.visit = visit_name
        db.add(session)
        db.commit()
    db.close()
    return client_info


@router.get("/sessions")
async def get_sessions(db=murfey_db):
    sessions = db.exec(select(Session)).all()
    clients = db.exec(select(ClientEnvironment)).all()
    res = []
    for sess in sessions:
        r = {"session": sess, "clients": []}
        for cl in clients:
            if cl.session_id == sess.id:
                r["clients"].append(cl)
        res.append(r)
    return res


@router.delete("/sessions/{session_id}")
def remove_session(session_id: MurfeySessionID, db=murfey_db):
    remove_session_by_id(session_id, db)


@router.get("/sessions/{session_id}/data_collection_groups")
def get_dc_groups(
    session_id: MurfeySessionID, db=murfey_db
) -> Dict[str, DataCollectionGroup]:
    data_collection_groups = db.exec(
        select(DataCollectionGroup).where(DataCollectionGroup.session_id == session_id)
    ).all()
    return {dcg.tag: dcg for dcg in data_collection_groups}


@router.post("/sessions/{session_id}/successful_processing")
def register_processing_success_in_ispyb(
    session_id: MurfeySessionID, db=ispyb_db, murfey_db=murfey_db
):
    collected_ids = murfey_db.exec(
        select(DataCollectionGroup, DataCollection, ProcessingJob, AutoProcProgram)
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollection.dcg_id == DataCollectionGroup.id)
        .where(ProcessingJob.dc_id == DataCollection.id)
        .where(AutoProcProgram.pj_id == ProcessingJob.id)
    ).all()
    appids = [c[3].id for c in collected_ids]
    if _transport_object:
        if db is not None:
            apps = db.query(ISPyBAutoProcProgram).filter(
                ISPyBAutoProcProgram.autoProcProgramId.in_(appids)
            )
            for updated in apps:
                updated.processingStatus = True
                _transport_object.do_update_processing_status(updated)


@router.get("/num_movies")
def count_number_of_movies(db=murfey_db) -> Dict[str, int]:
    res = db.exec(
        select(Movie.tag, func.count(Movie.murfey_id)).group_by(Movie.tag)
    ).all()
    return {r[0]: r[1] for r in res}


class PostInfo(BaseModel):
    router_name: str
    function_name: str
    data: dict
    kwargs: dict


@router.post("/instruments/{instrument_name}/failed_client_post")
def failed_client_post(instrument_name: str, post_info: PostInfo):
    zocalo_message = {
        "register": "failed_client_post",
        "router_name": post_info.router_name,
        "function_name": post_info.function_name,
        "data": post_info.data,
        "kwargs": post_info.kwargs,
    }
    if _transport_object:
        _transport_object.send(_transport_object.feedback_queue, zocalo_message)


@router.post("/sessions/{session_id}/rsyncer")
def register_rsyncer(session_id: int, rsyncer_info: RsyncerInfo, db=murfey_db):
    visit_name = db.exec(select(Session).where(Session.id == session_id)).one().visit
    rsync_instance = RsyncInstance(
        source=rsyncer_info.source,
        session_id=rsyncer_info.session_id,
        transferring=rsyncer_info.transferring,
        destination=rsyncer_info.destination,
        tag=rsyncer_info.tag,
    )
    db.add(rsync_instance)
    db.commit()
    db.close()
    prom.seen_files.labels(rsync_source=rsyncer_info.source, visit=visit_name).set(0)
    prom.transferred_files.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).set(0)
    prom.transferred_files_bytes.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).set(0)
    prom.seen_data_files.labels(rsync_source=rsyncer_info.source, visit=visit_name).set(
        0
    )
    prom.transferred_data_files.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).set(0)
    prom.transferred_data_files_bytes.labels(
        rsync_source=rsyncer_info.source, visit=visit_name
    ).set(0)
    prom.skipped_files.labels(rsync_source=rsyncer_info.source, visit=visit_name).set(0)
    return rsyncer_info


@router.get("/sessions/{session_id}/rsyncers", response_model=List[RsyncInstance])
def get_rsyncers_for_session(session_id: MurfeySessionID, db=murfey_db):
    rsync_instances = db.exec(
        select(RsyncInstance).where(RsyncInstance.session_id == session_id)
    )
    return rsync_instances.all()


class StringOfPathModel(BaseModel):
    path: str


@router.post("/sessions/{session_id}/rsyncer_stopped")
def register_stopped_rsyncer(
    session_id: int, rsyncer_source: StringOfPathModel, db=murfey_db
):
    rsyncer = db.exec(
        select(RsyncInstance)
        .where(RsyncInstance.session_id == session_id)
        .where(RsyncInstance.source == rsyncer_source.path)
    ).one()
    rsyncer.transferring = False
    db.add(rsyncer)
    db.commit()


@router.post("/sessions/{session_id}/rsyncer_started")
def register_restarted_rsyncer(
    session_id: int, rsyncer_source: StringOfPathModel, db=murfey_db
):
    rsyncer = db.exec(
        select(RsyncInstance)
        .where(RsyncInstance.session_id == session_id)
        .where(RsyncInstance.source == rsyncer_source.path)
    ).one()
    rsyncer.transferring = True
    db.add(rsyncer)
    db.commit()


@router.delete("/sessions/{session_id}/rsyncer")
def delete_rsyncer(session_id: int, rsyncer_source: StringOfPathModel, db=murfey_db):
    try:
        rsync_instance = db.exec(
            select(RsyncInstance)
            .where(RsyncInstance.session_id == session_id)
            .where(RsyncInstance.source == str(rsyncer_source.path))
        ).one()
        db.delete(rsync_instance)
        db.commit()
    except Exception:
        logger.error(
            f"Failed to delete rsyncer for source directory {sanitise(str(rsyncer_source.path))!r} "
            f"in session {session_id}.",
            exc_info=True,
        )


spa_router = APIRouter(
    prefix="/session_control/spa",
    dependencies=[Depends(validate_instrument_token)],
    tags=["Session Control: SPA"],
)


@spa_router.get("/sessions/{session_id}/grid_squares")
def get_grid_squares(session_id: MurfeySessionID, db=murfey_db):
    return _get_grid_squares(session_id, db)


@spa_router.get("/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares")
def get_grid_squares_from_dcg(
    session_id: MurfeySessionID, dcgid: int, db=murfey_db
) -> List[GridSquare]:
    return _get_grid_squares_from_dcg(session_id, dcgid, db)


@spa_router.get(
    "/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares/{gsid}/foil_holes"
)
def get_foil_holes_from_grid_square(
    session_id: MurfeySessionID, dcgid: int, gsid: int, db=murfey_db
) -> List[FoilHole]:
    return _get_foil_holes_from_grid_square(session_id, dcgid, gsid, db)


@spa_router.get("/sessions/{session_id}/foil_hole/{fh_name}")
def get_foil_hole(
    session_id: MurfeySessionID, fh_name: int, db=murfey_db
) -> Dict[str, int]:
    return _get_foil_hole(session_id, fh_name, db)


class AtlasRegistration(BaseModel):
    name: str
    acqusition_uuid: str


@spa_router.post("/sessions/{session_id}/register_atlas")
def register_atlas(
    session_id: MurfeySessionID,
    atlas_registration_data: AtlasRegistration,
    db=murfey_db,
):
    if SMARTEM_ACTIVE:
        session = db.exec(select(Session).where(Session.id == session_id)).one()
        machine_config = get_machine_config(session.instrument_name)[
            session.instrument_name
        ]
        if machine_config.smartem_api_url:
            smartem_client = SmartEMAPIClient(
                base_url=machine_config.smartem_api_url, logger=logger
            )
            possible_grids = smartem_client.get_acquisition_grids(
                atlas_registration_data.acqusition_uuid
            )
            grid_uuid = None
            for grid in possible_grids:
                if grid.name == atlas_registration_data.name.replace("_atlas", ""):
                    grid_uuid = grid.uuid
                    break
            if grid_uuid is not None:
                atlas_data = AtlasData(
                    id=atlas_registration_data.name,
                    acquisition_data=datetime.now(),
                    storage_folder="",
                    name=atlas_registration_data.name,
                    tiles=[],
                    gridsquare_positions=None,
                    grid_uuid=grid_uuid,
                )
                smartem_client.create_grid_atlas(atlas_data)
    else:
        logger.info("smartem deactivated so did not register atlas")


@spa_router.post("/sessions/{session_id}/make_atlas_jpg")
def make_atlas_jpg(
    session_id: MurfeySessionID, atlas_mrc: StringOfPathModel, db=murfey_db
):
    logger.debug(
        f"Received request to create JPG image of atlas {sanitise(atlas_mrc.path)!r}"
    )
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    return atlas_jpg_from_mrc(
        session.instrument_name, session.visit, Path(atlas_mrc.path)
    )


@spa_router.post("/sessions/{session_id}/grid_square/{gsid}")
def register_grid_square(
    session_id: MurfeySessionID,
    gsid: int,
    grid_square_params: GridSquareParameters,
    db=murfey_db,
):
    return _register_grid_square(session_id, gsid, grid_square_params, db)


@spa_router.post("/sessions/{session_id}/grid_square/{gs_name}/foil_hole")
def register_foil_hole(
    session_id: MurfeySessionID,
    gs_name: int,
    foil_hole_params: FoilHoleParameters,
    db=murfey_db,
):
    logger.info(
        f"Registering foil hole {foil_hole_params.name} with position {(foil_hole_params.x_location, foil_hole_params.y_location)}"
    )
    return _register_foil_hole(session_id, gs_name, foil_hole_params, db)


tomo_router = APIRouter(
    prefix="/session_control/tomo",
    dependencies=[Depends(validate_instrument_token)],
    tags=["Session Control: CryoET"],
)


@tomo_router.post("/sessions/{session_id}/search_map/{sm_name}")
def register_search_map(
    session_id: MurfeySessionID,
    sm_name: str,
    search_map_params: SearchMapParameters,
    db=murfey_db,
):
    return register_search_map_in_database(session_id, sm_name, search_map_params, db)


@tomo_router.post("/sessions/{session_id}/batch_position/{batch_name}")
def register_batch_position(
    session_id: MurfeySessionID,
    batch_name: str,
    batch_params: BatchPositionParameters,
    db=murfey_db,
):
    return register_batch_position_in_database(session_id, batch_name, batch_params, db)


correlative_router = APIRouter(
    prefix="/session_control/correlative",
    dependencies=[Depends(validate_instrument_token)],
    tags=["Session Control: Correlative Imaging"],
)


@correlative_router.get("/sessions/{session_id}/upstream_visits")
async def find_upstream_visits(session_id: MurfeySessionID, db=murfey_db):
    return _find_upstream_visits(session_id=session_id, db=db)


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_file_paths"
)
async def gather_upstream_files(
    visit_name: str,
    session_id: MurfeySessionID,
    upstream_file_request: UpstreamFileRequestInfo,
    db=murfey_db,
):
    return _gather_upstream_files(
        session_id=session_id,
        upstream_instrument=upstream_file_request.upstream_instrument,
        upstream_visit_path=upstream_file_request.upstream_visit_path,
        db=db,
    )


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_file/{upstream_file_path:path}"
)
async def get_upstream_file(
    visit_name: str,
    session_id: MurfeySessionID,
    upstream_file_path: str,
    db=murfey_db,
):
    upstream_file = _get_upstream_file(upstream_file_path)
    return (
        FileResponse(path=upstream_file) if upstream_file is not None else upstream_file
    )


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_tiff_paths"
)
async def gather_upstream_tiffs(visit_name: str, session_id: int, db=murfey_db):
    return _gather_upstream_tiffs(visit_name=visit_name, session_id=session_id, db=db)


@correlative_router.get(
    "/visits/{visit_name}/sessions/{session_id}/upstream_tiff/{tiff_path:path}"
)
async def get_tiff_file(visit_name: str, session_id: int, tiff_path: str, db=murfey_db):
    tiff_file = _get_tiff_file(
        visit_name=visit_name, session_id=session_id, tiff_path=tiff_path, db=db
    )
    return FileResponse(path=tiff_file) if tiff_file is not None else tiff_file



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from murfey.server.api.session_shared import find_upstream_visits, gather_upstream_files
from murfey.util.config import MachineConfig
from tests.conftest import ExampleVisit


@pytest.mark.parametrize("recurse", (True, False))
def test_find_upstream_visits(
    mocker: MockerFixture,
    tmp_path: Path,
    recurse: bool,
):
    # Get the visit, instrument name, and session ID
    visit_name_root = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}"
    visit_name = f"{visit_name_root}-{ExampleVisit.visit_number}"
    instrument_name = ExampleVisit.instrument_name
    session_id = ExampleVisit.murfey_session_id

    # Mock the database call
    mock_murfey_session_row = MagicMock()
    mock_murfey_session_row.visit = visit_name
    mock_murfey_session_row.instrument_name = instrument_name
    mock_murfey_db = MagicMock()
    mock_murfey_db.exec.return_value.one.return_value = mock_murfey_session_row

    # Create mock upstream visit directories and necessary data structures
    upstream_visits = {}
    upstream_data_dirs = {}
    for n in range(10):
        upstream_instrument = f"{instrument_name}{str(n).zfill(2)}"
        # Create path to visit
        upstream_visit = (
            tmp_path / f"{upstream_instrument}/data/2020/{visit_name_root}-{n}"
        )
        # Create some as directories, and some as files
        if n % 2:
            # Only directories should be picked up
            upstream_visit.mkdir(parents=True, exist_ok=True)
            upstream_visits[upstream_instrument] = {upstream_visit.stem: upstream_visit}
            # Check that the function can cope with recursive searching
            upstream_data_dirs[upstream_instrument] = (
                upstream_visit.parent.parent if recurse else upstream_visit.parent
            )
        else:
            upstream_visit.parent.mkdir(parents=True, exist_ok=True)
            upstream_visit.touch(exist_ok=True)

        # Create junk directories with multiple levels to test recursion logic with
        junk_directories = [
            tmp_path / f"{upstream_instrument}/data/junk/directory/number/{n}"
            for n in range(5)
        ]
        for dirpath in junk_directories:
            dirpath.mkdir(parents=True, exist_ok=True)

    # Mock the MachineConfig for this instrument
    mock_machine_config = MagicMock(spec=MachineConfig)
    mock_machine_config.upstream_data_directories = upstream_data_dirs
    mock_get_machine_config = mocker.patch(
        "murfey.server.api.session_shared.get_machine_config",
    )
    mock_get_machine_config.return_value = {instrument_name: mock_machine_config}

    # Run the function
    result = find_upstream_visits(session_id=session_id, db=mock_murfey_db)

    # Check that the expected directories are returned
    assert result == upstream_visits


def test_find_upstream_visits_permission_error(
    mocker: MockerFixture,
    tmp_path: Path,
):
    # Get the visit, instrument name, and session ID
    visit_name_root = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}"
    visit_name = f"{visit_name_root}-{ExampleVisit.visit_number}"
    instrument_name = ExampleVisit.instrument_name
    session_id = ExampleVisit.murfey_session_id

    # Mock the database call
    mock_murfey_session_row = MagicMock()
    mock_murfey_session_row.visit = visit_name
    mock_murfey_session_row.instrument_name = instrument_name
    mock_murfey_db = MagicMock()
    mock_murfey_db.exec.return_value.one.return_value = mock_murfey_session_row

    # Create mock upstream visit directories and necessary data structures
    upstream_visits: dict[str, dict] = {}
    upstream_data_dirs = {}
    for n in range(10):
        upstream_instrument = f"{instrument_name}{str(n).zfill(2)}"
        upstream_visit = (
            tmp_path / f"{upstream_instrument}/data/2020/{visit_name_root}-{n}"
        )
        # Create some as directories, and some as files
        if n % 2:
            upstream_visit.mkdir(parents=True, exist_ok=True)
            upstream_data_dirs[upstream_instrument] = upstream_visit.parent.parent
            # With os.scandir set to raise PermissionError, dictionaries should be empty
            upstream_visits[upstream_instrument] = {}
        else:
            upstream_visit.parent.mkdir(parents=True, exist_ok=True)
            upstream_visit.touch(exist_ok=True)

    # Mock the MachineConfig for this instrument
    mock_machine_config = MagicMock(spec=MachineConfig)
    mock_machine_config.upstream_data_directories = upstream_data_dirs
    mock_get_machine_config = mocker.patch(
        "murfey.server.api.session_shared.get_machine_config",
    )
    mock_get_machine_config.return_value = {instrument_name: mock_machine_config}

    # Mock the 'os.scandir' function used
    mocker.patch(
        "murfey.server.api.session_shared.os.scandir",
        side_effect=PermissionError(),
    )

    # Run the function:
    result = find_upstream_visits(session_id=session_id, db=mock_murfey_db)

    # With 'os.scandir' mocked to raise PermissionError, no entries should be returned
    assert result == upstream_visits


gather_upstream_files_test_matrix: tuple[
    tuple[tuple[str, list[str], list[str]], ...], ...
] = (
    # CLEM
    (
        # Search strings, files to match, and files to avoid
        (
            "processed/**/composite*.tiff",
            [
                file
                for sublist in [
                    [
                        f"processed/grid1/TileScan1/Position_{n}/composite_BF_FL.tiff"
                        for n in range(5)
                    ],
                ]
                for file in sublist
            ],
            [
                file
                for sublist in [
                    [
                        f"processed/grid1/TileScan1/Position_{n}/{color}.tiff"
                        for n in range(5)
                        for color in ("gray", "green", "red")
                    ],
                ]
                for file in sublist
            ],
        ),
        (
            "screenshots/**/*",
            [
                file
                for sublist in [
                    [f"screenshots/overview_{n}.png" for n in range(10)],
                    [f"screenshots/annotated_{n}.png" for n in range(10)],
                ]
                for file in sublist
            ],
            [],
        ),
    ),
    # FIB
    (
        # Search strings, files to match, and files to avoid
        (
            "maps/**/*",
            [
                file
                for sublist in [
                    [f"maps/data_{n}.txt" for n in range(5)],
                    [f"maps/map/image_{n}.tiff" for n in range(5)],
                ]
                for file in sublist
            ],
            [],
        ),
    ),
)


@pytest.mark.parametrize("test_params", gather_upstream_files_test_matrix)
def test_gather_upstream_files(
    mocker: MockerFixture,
    tmp_path: Path,
    test_params: tuple[tuple[str, list[str], list[str]], ...],
):
    # Get the visit, instrument name, and session ID
    visit_name_root = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}"
    visit_name = f"{visit_name_root}-{ExampleVisit.visit_number}"
    instrument_name = ExampleVisit.instrument_name
    session_id = ExampleVisit.murfey_session_id

    # Unpack the test params
    search_strings = [item[0] for item in test_params]
    upstream_relative_paths = [file for item in test_params for file in item[1]]
    other_relative_paths = [file for item in test_params for file in item[2]]

    # Set the upstream instrument and upstream visit to access
    upstream_instrument = f"{instrument_name}01"
    upstream_visit = f"{visit_name_root}-5"
    upstream_visit_path = tmp_path / f"{upstream_instrument}/data/2020/{upstream_visit}"

    # Construct the files and directories
    upstream_files = [
        upstream_visit_path / relative_path for relative_path in upstream_relative_paths
    ]
    other_files = [
        upstream_visit_path / relative_path for relative_path in other_relative_paths
    ]

    for file in upstream_files:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        file.touch(exist_ok=True)
        assert file.is_file()
    for file in other_files:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        file.touch(exist_ok=True)
        assert file.is_file()

    # Mock the database call
    mock_murfey_session_row = MagicMock()
    mock_murfey_session_row.visit = visit_name
    mock_murfey_session_row.instrument_name = instrument_name
    mock_murfey_db = MagicMock()
    mock_murfey_db.exec.return_value.one.return_value = mock_murfey_session_row

    # Mock the MachineConfig for this instrument
    mock_machine_config = MagicMock(spec=MachineConfig)
    mock_machine_config.upstream_data_search_strings = {
        upstream_instrument: search_strings
    }
    mock_get_machine_config = mocker.patch(
        "murfey.server.api.session_shared.get_machine_config",
    )
    mock_get_machine_config.return_value = {instrument_name: mock_machine_config}

    assert sorted(
        gather_upstream_files(
            session_id=session_id,
            upstream_instrument=upstream_instrument,
            upstream_visit_path=upstream_visit_path,
            db=mock_murfey_db,
        )
    ) == sorted(upstream_files)



import logging
import time

import ispyb.sqlalchemy._auto_db_schema as ISPyBDB
from sqlmodel.orm.session import Session as SQLModelSession

from murfey.server import _transport_object

logger = logging.getLogger("murfey.workflows.register_data_collection_group")


def run(message: dict, murfey_db: SQLModelSession) -> dict[str, bool]:
    # Fail immediately if no transport wrapper is found
    if _transport_object is None:
        logger.error("Unable to find transport manager")
        return {"success": False, "requeue": False}

    logger.info(f"Updating the experiment type for data collection group: \n{message}")

    record = ISPyBDB.DataCollectionGroup(
        dataCollectionGroupId=message["dcgid"],
        experimentTypeId=message["experiment_type_id"],
    )
    dcgid = _transport_object.do_update_data_collection_group(record).get(
        "return_value", None
    )

    if dcgid is None:
        time.sleep(2)
        logger.error(
            "Failed to update the following data collection group: \n"
            f"{message} \n"
            "Requeuing message"
        )
        return {"success": False, "requeue": True}

    return {"success": True}



import sys
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

import murfey
from murfey.cli.generate_openapi_schema import run

params_matrix: tuple[tuple[str | None, str | None, bool], ...] = (
    # Target | Output | To File
    (None, None, False),
    ("instrument-server", "json", True),
    ("server", "yaml", False),
    ("instrument-server", "yaml", False),
    ("server", "json", True),
)


@pytest.mark.parametrize("test_params", params_matrix)
def test_run(
    mocker: MockerFixture,
    tmp_path: Path,
    test_params: tuple[str | None, str | None, bool],
):
    # Unpack test params
    target, output, to_file = test_params

    # Mock out print() and exit()
    mock_print = mocker.patch("builtins.print")
    mock_exit = mocker.patch("builtins.exit")

    # Construct the CLI args
    sys_args = [""]
    if target is not None:
        sys_args.extend(["-t", target])
    if output is not None:
        sys_args.extend(["-o", output])

    target = target if target is not None else "server"
    output = output if output is not None else "yaml"
    if to_file:
        save_path = tmp_path / f"openapi.{output}"
        sys_args.extend(["-f", str(save_path)])
    else:
        save_path = Path(murfey.__path__[0]) / "util" / f"openapi-{target}.{output}"
    sys_args.extend(["--debug"])
    sys.argv = sys_args

    # Run the function and check that it runs as expected
    run()
    print_calls = mock_print.call_args_list
    last_print_call = print_calls[-1]
    last_printed = last_print_call.args[0]
    assert last_printed.startswith("OpenAPI schema saved to")
    mock_exit.assert_called_once()
    assert save_path.exists()


failure_params_matrix = (
    ["-t", "blah"],
    ["-o", "blah"],
)


@pytest.mark.parametrize("test_params", failure_params_matrix)
def test_run_fails(test_params: list[str]):
    # Construct the CLI args
    sys_args = [""]
    sys_args.extend(test_params)
    sys.argv = sys_args

    with pytest.raises(ValueError):
        run()



from unittest import mock
from unittest.mock import MagicMock

from pytest_mock import MockerFixture

from murfey.workflows.register_atlas_update import run


def test_run(
    mocker: MockerFixture,
):
    # Set up mocks and the dummy message to be registered
    mock_transport_object = mocker.patch(
        "murfey.workflows.register_atlas_update._transport_object"
    )
    mock_murfey_db = MagicMock()
    message = {
        "register": "atlas_update",
        "atlas_id": mock.sentinel,
        "atlas": mock.sentinel,
        "atlas_pixel_size": mock.sentinel,
        "sample": mock.sentinel,
    }

    # Run the function and check the results and calls made
    result = run(message, mock_murfey_db)
    mock_transport_object.do_update_atlas.assert_called_once_with(
        atlas_id=message["atlas_id"],
        atlas_image=message["atlas"],
        pixel_size=message["atlas_pixel_size"],
        slot=message["sample"],
        collection_mode=message.get("collection_mode"),
        color_flags=message.get("color_flags", {}),
    )
    assert result == {"success": True}



import json
import logging
from datetime import datetime
from typing import Any

from fastapi import APIRouter, Request

logger = logging.getLogger("murfey.server.api.logging")

router = APIRouter(
    prefix="/logging",
    tags=["Logging"],
)


@router.post("/logs")
async def forward_logs(request: Request):
    """
    Receives a list of stringified JSON log records from the instrument server,
    unpacks them, and forwards them through the handlers set up on the backend.
    """

    data: list[str] = await request.json()
    for line in data:
        log_data: dict[str, Any] = json.loads(line)
        logger_name = log_data["name"]
        log_data.pop("msecs", None)
        log_data.pop("relativeCreated", None)
        client_timestamp = log_data.pop("created", 0)
        if client_timestamp:
            log_data["client_time"] = datetime.fromtimestamp(
                client_timestamp
            ).isoformat()
        log_data["client_host"] = request.client.host if request.client else None
        logging.getLogger(logger_name).handle(logging.makeLogRecord(log_data))



from importlib.metadata import entry_points
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

feedback_callback_params_matrix = (
    # Murfey workflows currently present in pyproject.toml
    ("atlas_update",),
    ("clem.align_and_merge",),
    ("clem.process_raw_lifs",),
    ("clem.process_raw_tiffs",),
    ("clem.register_align_and_merge_result",),
    ("clem.register_preprocessing_result",),
    ("data_collection",),
    ("data_collection_group",),
    ("pato",),
    ("picked_particles",),
    ("picked_tomogram",),
    ("processing_job",),
    ("spa.flush_spa_preprocess",),
)


@pytest.mark.parametrize("test_params", feedback_callback_params_matrix)
def test_feedback_callback(
    mocker: MockerFixture,
    test_params: tuple[str],
):
    """
    Checks that feedback-callback loop works correctly for the entry points-based workflows
    """

    # Unpack test params
    (entry_point_name,) = test_params

    # Patch the functions used to generate the module-level variables
    mock_get_security_config = mocker.patch("murfey.util.config.get_security_config")
    mock_get_security_config.return_value = MagicMock()
    mock_url = mocker.patch("murfey.server.murfey_db.url")
    mock_url.return_value = MagicMock()
    mock_create_engine = mocker.patch("sqlmodel.create_engine")
    mock_create_engine.return_value = MagicMock()
    mock_murfey_db = MagicMock()
    mock_sql_session = mocker.patch("sqlmodel.Session")
    mock_sql_session.return_value = mock_murfey_db

    # Load the entry point and patch the executable it calls
    eps = list(entry_points(group="murfey.workflows", name=entry_point_name))
    assert len(eps) == 1  # Entry point should be present and unique
    mock_function = mocker.patch(eps[0].value.replace(":", "."))

    # Initialise after mocking
    from murfey.server.feedback import feedback_callback

    # Run the function and check that it calls the entry point correctly
    header = {"dummy": "dummy"}
    message = {"register": entry_point_name}
    feedback_callback(header, message, mock_murfey_db)
    mock_function.assert_called_once_with(message=message, murfey_db=mock_murfey_db)



import logging
from pathlib import Path

import mrcfile
import PIL.Image
from werkzeug.utils import secure_filename

from murfey.util import sanitise
from murfey.util.config import get_machine_config

logger = logging.getLogger("murfey.workflows.spa.atlas")


def atlas_jpg_from_mrc(instrument_name: str, visit_name: str, atlas_mrc: Path):
    logger.debug(
        f"Starting workflow to create JPG image of atlas {sanitise(str(atlas_mrc))!r}"
    )
    with mrcfile.open(atlas_mrc) as mrc:
        data = mrc.data

    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]

    parts = [secure_filename(p) for p in atlas_mrc.parts]
    visit_idx = parts.index(visit_name)
    core = Path("/".join(parts[: visit_idx + 1]))
    sample_id = "Sample"
    for p in parts:
        if "Sample" in p:
            sample_id = p
            break
    atlas_jpg_file = (
        core
        / machine_config.processed_directory_name
        / "atlas"
        / secure_filename(f"{sample_id}_{atlas_mrc.stem}_fullres.jpg")
    )
    atlas_jpg_file.parent.mkdir(parents=True, exist_ok=True)

    data = data - data.min()
    data = data.astype(float) * 255 / data.max()
    data = data.astype("uint8")
    PIL.Image.fromarray(data).save(atlas_jpg_file)
    logger.debug(f"JPG image of atlas saved as {str(atlas_jpg_file)!r}")



from __future__ import annotations

import re
import traceback
from ast import literal_eval
from importlib.metadata import (
    EntryPoint,  # type hinting only
    entry_points,
)
from logging import getLogger
from pathlib import Path
from typing import Literal, Optional, Type, Union

from fastapi import APIRouter
from pydantic import BaseModel, field_validator
from sqlalchemy.exc import NoResultFound
from sqlmodel import Session, select

from murfey.server import _transport_object
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise
from murfey.util.config import get_machine_config
from murfey.util.db import (
    CLEMImageMetadata,
    CLEMImageSeries,
    CLEMImageStack,
    CLEMLIFFile,
    CLEMTIFFFile,
    Session as MurfeySession,
)

# Set up logger
logger = getLogger("murfey.server.api.clem")

# Create APIRouter class object
router = APIRouter(
    prefix="/workflow/clem",
    tags=["Workflows: CLEM"],
)

# Valid file types
valid_file_types = (
    ".lif",
    ".tif",
    ".tiff",
    ".xlif",
    ".xml",
)


"""
HELPER FUNCTIONS
"""


def validate_and_sanitise(
    file: Path,
    session_id: int,
    db: Session,
) -> Path:
    """
    Performs validation and sanitisation on the incoming file paths, ensuring that
    no forbidden characters are present and that the the path points only to allowed
    sections of the file server.

    Returns the file path as a sanitised string that can be converted into a Path
    object again.

    NOTE: Due to the instrument name query, 'db' now needs to be passed as an
    explicit variable to this function from within a FastAPI endpoint, as using the
    instance that was imported directly won't load it in the correct state.
    """

    # Resolve symlinks and directory changes to get full file path
    full_path = Path(file).resolve()

    # Use machine configuration to validate which file base paths are accepted from
    instrument_name = (
        db.exec(select(MurfeySession).where(MurfeySession.id == session_id))
        .one()
        .instrument_name
    )
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    rsync_basepath = (machine_config.rsync_basepath or Path("")).resolve()

    # Check that full file path doesn't contain unallowed characters
    #   Currently allows only:
    #   - words (alphanumerics and "_"; \w),
    #   - spaces (\s),
    #   - periods,
    #   - dashes,
    #   - forward slashes ("/")
    if bool(re.fullmatch(r"^[\w\s\.\-/]+$", str(full_path))) is False:
        raise ValueError(f"Unallowed characters present in {file}")

    # Check that it's not accessing somehwere it's not allowed
    if not str(full_path).startswith(str(rsync_basepath)):
        raise ValueError(f"{file} points to a directory that is not permitted")

    # Check that it is of a permitted file type
    if f"{full_path.suffix}" not in valid_file_types:
        raise ValueError(f"{full_path.suffix} is not a permitted file format")

    return full_path


def get_db_entry(
    db: Session,
    # With the database search funcion having been moved out of the FastAPI
    # endpoint, the database now has to be explicitly passed within the FastAPI
    # endpoint function in order for it to be loaded in the correct state.
    table: Type[
        Union[
            CLEMImageMetadata,
            CLEMImageSeries,
            CLEMImageStack,
            CLEMLIFFile,
            CLEMTIFFFile,
        ]
    ],
    session_id: int,
    file_path: Optional[Path] = None,
    series_name: Optional[str] = None,
) -> Union[
    CLEMImageMetadata,
    CLEMImageSeries,
    CLEMImageStack,
    CLEMLIFFile,
    CLEMTIFFFile,
]:
    """
    Searches the CLEM workflow-related tables in the Murfey database for an entry that
    matches the file path or series name within a given session. Returns the entry if
    a match is found, otherwise register it as a new entry in the database.
    """

    # Validate that parameters are provided correctly
    if file_path is None and series_name is None:
        raise ValueError(
            "One of either 'file_path' or 'series_name' has to be provided"
        )
    if file_path is not None and series_name is not None:
        raise ValueError("Only one of 'file_path' or 'series_name' should be provided")

    # Validate file path if provided
    if file_path is not None:
        try:
            file_path = validate_and_sanitise(file_path, session_id, db)
        except Exception:
            raise Exception

    # Validate series name to use
    if series_name is not None:
        if bool(re.fullmatch(r"^[\w\s\.\-/]+$", series_name)) is False:
            raise ValueError("One or more characters in the string are not permitted")

    # Return database entry if it exists
    try:
        db_entry = (
            db.exec(
                select(table)
                .where(table.session_id == session_id)
                .where(table.file_path == str(file_path))
            ).one()
            if file_path is not None
            else db.exec(
                select(table)
                .where(table.session_id == session_id)
                .where(table.series_name == series_name)
            ).one()
        )
    # Create and register new entry if not present
    except NoResultFound:
        db_entry = (
            table(
                file_path=str(file_path),
                session_id=session_id,
            )
            if file_path is not None
            else table(
                series_name=series_name,
                session_id=session_id,
            )
        )
        db.add(db_entry)
        db.commit()

    except Exception:
        raise Exception

    return db_entry


"""
API ENDPOINTS FOR FILE REGISTRATION
"""


class LifInfo(BaseModel):
    lif_file: Path
    master_metadata: Optional[Path] = None
    child_metadata: list[Path] = []
    child_series: list[str] = []
    child_stacks: list[Path] = []


@router.post("/sessions/{session_id}/clem/lif_files")
def register_lif_file(
    lif_file: LifInfo,
    session_id: int,
    db: Session = murfey_db,
):
    # Return or register the LIF file entry
    try:
        clem_lif_file: CLEMLIFFile = get_db_entry(
            db=db,
            table=CLEMLIFFile,
            session_id=session_id,
            file_path=lif_file.lif_file,
        )
    except Exception:
        logger.error(
            "Exception encountered while registering "
            f"LIF file {sanitise(str(lif_file))!r}: \n"
            f"{traceback.format_exc()}"
        )
        return False

    # Add metadata information if provided
    if lif_file.master_metadata is not None:
        try:
            master_metadata = validate_and_sanitise(
                lif_file.master_metadata, session_id, db
            )
            clem_lif_file.master_metadata = str(master_metadata)
        except Exception:
            logger.warning(
                "Unable to add master metadata information to database entry for "
                f"LIF file {sanitise(str(lif_file))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register child metadata if provided
    for metadata in lif_file.child_metadata:
        try:
            metadata_db_entry: CLEMImageMetadata = get_db_entry(
                db=db,
                table=CLEMImageMetadata,
                session_id=session_id,
                file_path=metadata,
            )
            # Append to database entry
            clem_lif_file.child_metadata.append(metadata_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"metadata file {sanitise(str(metadata))!r} in association with "
                f"LIF file {sanitise(str(lif_file))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Register child image series if provided
    for series in lif_file.child_series:
        try:
            series_db_entry: CLEMImageSeries = get_db_entry(
                db=db,
                table=CLEMImageSeries,
                session_id=session_id,
                series_name=series,
            )
            # Append to database entry
            clem_lif_file.child_series.append(series_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"metadata file {sanitise(series)!r} in association with "
                f"LIF file {sanitise(str(lif_file))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Register child image stacks if provided
    for stack in lif_file.child_stacks:
        try:
            stack_db_entry: CLEMImageStack = get_db_entry(
                db=db,
                table=CLEMImageStack,
                session_id=session_id,
                file_path=stack,
            )
            # Append to database entry
            clem_lif_file.child_stacks.append(stack_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"image stack {sanitise(str(stack))!r} in association with "
                f"LIF file {sanitise(str(lif_file))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Commit to database
    db.add(clem_lif_file)
    db.commit()
    db.close()
    return True


class TiffInfo(BaseModel):
    tiff_file: Path
    associated_metadata: Optional[Path] = None
    associated_series: Optional[str] = None
    associated_stack: Optional[Path] = None


@router.post("/sessions/{session_id}/clem/tiff_files")
def register_tiff_file(
    tiff_file: TiffInfo,
    session_id: int,
    db: Session = murfey_db,
):
    # Get or register the database entry
    try:
        clem_tiff_file: CLEMTIFFFile = get_db_entry(
            db=db,
            table=CLEMTIFFFile,
            session_id=session_id,
            file_path=tiff_file.tiff_file,
        )
    except Exception:
        logger.error(
            "Exception encountered while registering "
            f"TIFF file {sanitise(str(tiff_file))!r}: \n"
            f"{traceback.format_exc()}"
        )
        return False

    # Add metadata if provided
    if tiff_file.associated_metadata is not None:
        try:
            metadata_db_entry: CLEMImageMetadata = get_db_entry(
                db=db,
                table=CLEMImageMetadata,
                session_id=session_id,
                file_path=tiff_file.associated_metadata,
            )
            # Link database entries
            clem_tiff_file.associated_metadata = metadata_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"metadata file {sanitise(str(tiff_file.associated_metadata))!r} in association with "
                f"TIFF file {sanitise(str(tiff_file))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Add series information if provided
    if tiff_file.associated_series is not None:
        try:
            series_db_entry: CLEMImageSeries = get_db_entry(
                db=db,
                table=CLEMImageSeries,
                session_id=session_id,
                series_name=tiff_file.associated_series,
            )
            # Link database entries
            clem_tiff_file.child_series = series_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"CLEM series {sanitise(tiff_file.associated_series)!r} in association with "
                f"TIFF file {sanitise(str(tiff_file))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Add image stack information if provided
    if tiff_file.associated_stack is not None:
        try:
            stack_db_entry: CLEMImageStack = get_db_entry(
                db=db,
                table=CLEMImageStack,
                session_id=session_id,
                file_path=tiff_file.associated_stack,
            )
            # Link database entries
            clem_tiff_file.child_stack = stack_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"image stack {sanitise(str(tiff_file.associated_stack))!r} in association with "
                f"{traceback.format_exc()}"
            )

    # Commit to database
    db.add(clem_tiff_file)
    db.commit()
    db.close()
    return True


@router.post("/sessions/{session_id}/clem/metadata_files")
def register_clem_metadata(
    metadata_file: Path,
    session_id: int,
    parent_lif: Optional[Path] = None,
    associated_tiffs: list[Path] = [],
    associated_series: Optional[str] = None,
    associated_stacks: list[Path] = [],
    db: Session = murfey_db,
):
    # Return database entry if it already exists
    try:
        clem_metadata: CLEMImageMetadata = get_db_entry(
            db=db,
            table=CLEMImageMetadata,
            session_id=session_id,
            file_path=metadata_file,
        )
    except Exception:
        logger.error(
            "Exception encountered while registering"
            f"metadata file {sanitise(str(metadata_file))!r}"
            f"{traceback.format_exc()}"
        )
        return False

    # Register a parent LIF file if provided
    if parent_lif is not None:
        try:
            lif_db_entry: CLEMLIFFile = get_db_entry(
                db=db,
                table=CLEMLIFFile,
                session_id=session_id,
                file_path=parent_lif,
            )
            # Link database entries
            clem_metadata.parent_lif = lif_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"LIF file {sanitise(str(parent_lif))!r} in association with "
                f"metadata file {sanitise(str(metadata_file))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register associated TIFF files if provided
    for tiff in associated_tiffs:
        try:
            tiff_db_entry: CLEMTIFFFile = get_db_entry(
                db=db,
                table=CLEMTIFFFile,
                session_id=session_id,
                file_path=tiff,
            )
            # Append entry
            clem_metadata.associated_tiffs.append(tiff_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"TIFF file {sanitise(str(tiff))!r} in association with "
                f"metadata file {sanitise(str(metadata_file))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Register associated image series if provided
    if associated_series is not None:
        try:
            series_db_entry: CLEMImageSeries = get_db_entry(
                db=db,
                table=CLEMImageSeries,
                session_id=session_id,
                series_name=associated_series,
            )
            # The link can only be made from series-side; not sure why
            series_db_entry.associated_metadata = clem_metadata
            db.add(series_db_entry)
            db.commit()
        except Exception:
            logger.warning(
                "Unable to register "
                f"CLEM series {sanitise(associated_series)!r} in association with "
                f"metadata file {sanitise(str(metadata_file))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register associated image stacks if provided
    for stack in associated_stacks:
        try:
            stack_db_entry: CLEMImageStack = get_db_entry(
                db=db,
                table=CLEMImageStack,
                session_id=session_id,
                file_path=stack,
            )
            clem_metadata.associated_stacks.append(stack_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"image stack {sanitise(str(stack))!r} in association with "
                f"metadata file {sanitise(str(metadata_file))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Commit to database
    db.add(clem_metadata)
    db.commit()
    db.close()
    return True


@router.post("/sessions/{session_id}/clem/image_series")
def register_image_series(
    series_name: str,
    session_id: int,
    parent_lif: Optional[Path] = None,
    parent_tiffs: list[Path] = [],
    associated_metadata: Optional[Path] = None,
    child_stacks: list[Path] = [],
    db: Session = murfey_db,
):
    # Get or register series
    try:
        clem_image_series: CLEMImageSeries = get_db_entry(
            db=db,
            table=CLEMImageSeries,
            session_id=session_id,
            series_name=series_name,
        )
    except Exception:
        logger.error(
            "Exception encountered while registering "
            f"CLEM series {sanitise(series_name)!r}: \n"
            f"{traceback.format_exc()}"
        )
        return False

    # Register parent LIF file if provided
    if parent_lif is not None:
        try:
            lif_db_entry: CLEMLIFFile = get_db_entry(
                db=db,
                table=CLEMLIFFile,
                session_id=session_id,
                file_path=parent_lif,
            )
            # Link entries
            clem_image_series.parent_lif = lif_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"LIF file {sanitise(str(parent_lif))!r} in association with "
                f"CLEM series {sanitise(series_name)!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register parent TIFFs if provided
    for tiff in parent_tiffs:
        try:
            tiff_db_entry: CLEMTIFFFile = get_db_entry(
                db=db,
                table=CLEMTIFFFile,
                session_id=session_id,
                file_path=tiff,
            )
            # Append entry
            clem_image_series.parent_tiffs.append(tiff_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"TIFF file {sanitise(str(tiff))!r} in association with "
                f"CLEM series {sanitise(series_name)!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue  # Try next item in loop

    # Register associated metadata if provided
    if associated_metadata is not None:
        try:
            metadata_db_entry: CLEMImageMetadata = get_db_entry(
                db=db,
                table=CLEMImageMetadata,
                session_id=session_id,
                file_path=associated_metadata,
            )
            # Link entries
            clem_image_series.associated_metadata = metadata_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"metadata file {sanitise(str(associated_metadata))!r} in association with "
                f"CLEM series {sanitise(series_name)!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register child image stacks if provided
    for stack in child_stacks:
        try:
            stack_db_entry: CLEMImageStack = get_db_entry(
                db=db,
                table=CLEMImageStack,
                session_id=session_id,
                file_path=stack,
            )
            # Append entry
            clem_image_series.child_stacks.append(stack_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"image stack {sanitise(str(stack))!r} in association with "
                f"CLEM series {sanitise(series_name)!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Register
    db.add(clem_image_series)
    db.commit()
    db.close()
    return True


@router.post("/sessions/{session_id}/clem/image_stacks")
def register_image_stack(
    image_stack: Path,
    session_id: int,
    channel: Optional[str] = None,
    parent_lif: Optional[Path] = None,
    parent_tiffs: list[Path] = [],
    associated_metadata: Optional[Path] = None,
    parent_series: Optional[str] = None,
    db: Session = murfey_db,
):
    # Get or register image stack entry
    try:
        clem_image_stack: CLEMImageStack = get_db_entry(
            db=db,
            table=CLEMImageStack,
            session_id=session_id,
            file_path=image_stack,
        )
    except Exception:
        logger.error(
            "Exception encountered while registering "
            f"image stack {sanitise(str(image_stack))!r}: \n"
            f"{traceback.format_exc()}"
        )
        return False

    # Register channel name if provided
    if channel is not None:
        clem_image_stack.channel_name = channel

    # Register parent LIF file if provided
    if parent_lif is not None:
        try:
            lif_db_entry: CLEMLIFFile = get_db_entry(
                db=db,
                table=CLEMLIFFile,
                session_id=session_id,
                file_path=parent_lif,
            )
            clem_image_stack.parent_lif = lif_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"LIF file {sanitise(str(parent_lif))!r} in association with "
                f"image stack {sanitise(str(image_stack))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register parent TIFF files if provided
    for tiff in parent_tiffs:
        try:
            tiff_db_entry: CLEMTIFFFile = get_db_entry(
                db=db,
                table=CLEMTIFFFile,
                session_id=session_id,
                file_path=tiff,
            )
            # Append entry
            clem_image_stack.parent_tiffs.append(tiff_db_entry)
        except Exception:
            logger.warning(
                "Unable to register "
                f"TIFF file {sanitise(str(tiff))!r} in association with "
                f"image stack {sanitise(str(image_stack))!r}: \n"
                f"{traceback.format_exc()}"
            )
            continue

    # Register associated metadata if provided
    if associated_metadata is not None:
        try:
            metadata_db_entry: CLEMImageMetadata = get_db_entry(
                db=db,
                table=CLEMImageMetadata,
                session_id=session_id,
                file_path=associated_metadata,
            )
            # Link entries
            clem_image_stack.associated_metadata = metadata_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"metadata file {sanitise(str(associated_metadata))!r} in association with "
                f"image stack {sanitise(str(image_stack))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register parent series if provided
    if parent_series is not None:
        try:
            series_db_entry: CLEMImageSeries = get_db_entry(
                db=db,
                table=CLEMImageSeries,
                session_id=session_id,
                series_name=parent_series,
            )
            # Link entries
            clem_image_stack.parent_series = series_db_entry
        except Exception:
            logger.warning(
                "Unable to register "
                f"CLEM series {sanitise(parent_series)!r} in association with "
                f"image stack {sanitise(str(image_stack))!r}: \n"
                f"{traceback.format_exc()}"
            )

    # Register updates to entry
    db.add(clem_image_stack)
    db.commit()
    db.close()
    return True


"""
API ENDPOINTS FOR FILE PROCESSING
"""


@router.post(
    "/sessions/{session_id}/clem/preprocessing/process_raw_lifs"
)  # API posts to this URL
def process_raw_lifs(
    session_id: int,
    lif_file: LifInfo,
    db: Session = murfey_db,
):
    try:
        # Try and load relevant Murfey workflow
        workflow: EntryPoint = list(
            entry_points(group="murfey.workflows", name="clem.process_raw_lifs")
        )[0]
    except IndexError:
        raise RuntimeError("The relevant Murfey workflow was not found")

    # Get instrument name from the database to load the correct config file
    session_row: MurfeySession = db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    instrument_name = session_row.instrument_name

    # Pass arguments along to the correct workflow
    workflow.load()(
        # Match the arguments found in murfey.workflows.clem.process_raw_lifs
        file=lif_file.lif_file,
        root_folder="images",
        session_id=session_id,
        instrument_name=instrument_name,
        messenger=_transport_object,
    )
    return True


class TIFFSeriesInfo(BaseModel):
    series_name: str
    tiff_files: list[Path]
    series_metadata: Path


@router.post("/sessions/{session_id}/clem/preprocessing/process_raw_tiffs")
def process_raw_tiffs(
    session_id: int,
    tiff_info: TIFFSeriesInfo,
    db: Session = murfey_db,
):
    try:
        # Try and load relevant Murfey workflow
        workflow: EntryPoint = list(
            entry_points(group="murfey.workflows", name="clem.process_raw_tiffs")
        )[0]
    except IndexError:
        raise RuntimeError("The relevant Murfey workflow was not found")

    # Get instrument name from the database to load the correct config file
    session_row: MurfeySession = db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    instrument_name = session_row.instrument_name

    # Pass arguments to correct workflow
    workflow.load()(
        # Match the arguments found in murfey.workflows.clem.process_raw_tiffs
        tiff_list=tiff_info.tiff_files,
        root_folder="images",
        session_id=session_id,
        instrument_name=instrument_name,
        metadata=tiff_info.series_metadata,
        messenger=_transport_object,
    )
    return True


class AlignAndMergeParams(BaseModel):
    # Processing parameters
    series_name: str
    images: list[Path]
    metadata: Path
    # Optional processing parameters
    crop_to_n_frames: Optional[int] = None
    align_self: Literal["enabled", ""] = ""
    flatten: Literal["mean", "min", "max", ""] = ""
    align_across: Literal["enabled", ""] = ""

    @field_validator("images", mode="before")
    @classmethod
    def parse_stringified_list(cls, value):
        if isinstance(value, str):
            try:
                eval_result = literal_eval(value)
                if isinstance(eval_result, list):
                    parent_tiffs = [Path(p) for p in eval_result]
                    return parent_tiffs
            except (SyntaxError, ValueError):
                raise ValueError("Unable to parse input")
        # Return value as-is; if it fails, it fails
        return value


@router.post("/sessions/{session_id}/clem/processing/align_and_merge_stacks")
def align_and_merge_stacks(
    session_id: int,
    align_and_merge_params: AlignAndMergeParams,
    db: Session = murfey_db,
):
    try:
        # Try and load relevant Murfey workflow
        workflow: EntryPoint = list(
            entry_points(group="murfey.workflows", name="clem.align_and_merge")
        )[0]
    except IndexError:
        raise RuntimeError("The relevant Murfey workflow was not found")

    # Get instrument name from the database to load the correct config file
    session_row: MurfeySession = db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    instrument_name = session_row.instrument_name

    # Pass arguments to correct workflow
    workflow.load()(
        # Match the arguments found in murfey.workflows.clem.align_and_merge
        # Session parameters
        session_id=session_id,
        instrument_name=instrument_name,
        # Processing parameters
        series_name=align_and_merge_params.series_name,
        images=align_and_merge_params.images,
        metadata=align_and_merge_params.metadata,
        # Optional processing parameters
        crop_to_n_frames=align_and_merge_params.crop_to_n_frames,
        align_self=align_and_merge_params.align_self,
        flatten=align_and_merge_params.flatten,
        align_across=align_and_merge_params.align_across,
        # Optional session parameters
        messenger=_transport_object,
    )
    return True
